import json
from pathlib import Path

from attrs import define


@define
class Settings:
    conn_str: str
    cities_data: str
    countries_data: str


def load_settings() -> Settings:
    with open(Path().resolve() / "src/gisterical/settings/settings.json", "r") as f:
        s = json.load(f)
        conn_str = f"postgresql+psycopg2://{s['user']}:{s['password']}@{s['hostname']}/{s['database_name']}"
        return Settings(conn_str=conn_str, cities_data=s['cities_data'], countries_data=s['countries_data'])
    

def update_settings():
    sett_dict = {}
    
    db_name = input('Enter database name: ')
    
    host = input('Enter hostname or Enter to keep default "localhost": ')
    if not host:
        host = 'localhost'
    
    user = input('Enter user name of database owner: ')
    password = input('Enter password of database owner: ')

    with open(Path().resolve() / "src/gisterical/settings/settings.json", "r") as f:
        s = json.load(f)
        
    sett_dict = {
        'cities_data': s['cities_data'],
        "countries_data": s["countries_data"],
        "database_name": db_name,
        "user": user,
        "password": password,
        "hostname": host,
    }
    
    with open(Path().resolve() / "src/gisterical/settings/settings.json", "w") as f:
        json.dump(sett_dict, f)
        

if __name__ == '__main__':
    update_settings()   

    


import datetime as dt
from pathlib import Path
from shutil import copyfile

from loguru import logger
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

from gisterical.database.schema import Image, Country, City
from gisterical.core.image_metadata import PhotoData
from gisterical.settings.settings import load_settings
from gisterical.util.file_meta import FileMeta


SETTINGS = load_settings()


class DbApi:
    conn_str = SETTINGS.conn_str
    engine: Engine = create_engine(conn_str)
    session = sessionmaker(engine)

    def add_photo_to_db(self, data: list[PhotoData]):
        logger.debug(f"Adding {len(data)} files to the database.")
        with self.session.begin() as sess:
            for d in data:
                if -999 not in {d.latitude, d.longitude, d.altitude}:
                    loc = f"POINTZ({d.longitude} {d.latitude} {d.altitude})"
                else:
                    loc = None
                acc = d.gps_accuracy if d.gps_accuracy != -999 else None
                direction = d.photo_direction if d.photo_direction != -999 else None

                new_result = Image(
                    path=d.path,
                    location=loc,
                    timestamp=d.timestamp,
                    gps_accuracy=acc,
                    photo_direction=direction,
                    device_make=d.camera_make,
                    device_model=d.camera_model,
                    phash=d.phash,
                    colorhash=d.colorhash,
                )

                sess.add(new_result)
                sess.flush()
            sess.commit()
        logger.debug("Successfully added!")

    def select_files_and_output(
        self,
        location: tuple[float, ...] = (0, 0, 0),
        distance_km: int | float = 1e10,
        start_date: dt.datetime = dt.datetime(1900, 1, 1),
        end_date: dt.datetime = dt.datetime(9999, 1, 1),
        target_folder: str | Path = "/home/pavel/Pictures/temp",
    ):
        logger.info(
            f"Searching for images within {distance_km} kilometers "
            f"from coordinates {location} between {start_date} "
            f"and {end_date}."
        )
        p = Path(target_folder)
        if not p.is_dir():
            p.mkdir()

        with self.session.begin() as sess:
            q = (
                sess.query(Image.path)
                .filter(
                    Image.location.ST_DistanceSphere(
                        f"POINTZ({location[0]} {location[1]} {location[2]})"
                    )
                    <= distance_km * 1000
                )
                .filter(Image.timestamp >= start_date)
                .filter(Image.timestamp <= end_date)
            )

            res: list[str] = [i[0] for i in q.all()]
            logger.info(f"{len(res)} files found. Copying the files to {target_folder}.")

        for pth in res:
            fname = Path(pth).name
            copyfile(str(pth), str(p / fname))
        logger.info("Success!")
        
    def get_photos_by_city(self, distance_km: int):
        logger.info('Querying images with nearest city data.')        
        with self.session.begin() as sess:
            q = sess.query(Image.path, Image.timestamp, City.name).distinct(Image.path).\
                join(City, Image.location.ST_DWithin(City.location, distance_km * 1000, True)).\
                    order_by(Image.path, City.population.desc()).all()
        return [FileMeta(path=Path(i[0]), date=i[1], city=i[2]) for i in q]
    
    def get_photo_path_date(self):
        logger.info("Querying photo datetime information")
        with self.session.begin() as sess:
            q = sess.query(Image.path, Image.timestamp).all()
            return [FileMeta(path=Path(i[0]), date=i[1]) for i in q]        
        
    def get_photo_city_country(self, distance_km: int):
        logger.info('Querying images with nearest city and country information.')
        with self.session.begin() as sess:
            q = sess.query(Image.path, Image.timestamp, Country.name, City.name).distinct(Image.path).\
                join(City, Image.location.ST_DWithin(City.location, distance_km * 1000, True)).\
                join(Country, Country.geometry.ST_Contains(Image.location)).\
                    order_by(Image.path, City.population.desc()).all()    
        return [FileMeta(path=Path(i[0]), date=i[1], country=i[2], city=i[3]) for i in q]
                    
    def get_photo_country(self):
        logger.info('Querying images with country information.')
        with self.session.begin() as sess:
            q = sess.query(Image.path, Image.timestamp, Country.name).\
                join(Country, Country.geometry.ST_Contains(Image.location)).all()
        return [FileMeta(path=Path(i[0]), date=i[1], country=i[2]) for i in q]   
    
    def get_photo_no_location(self):
        with self.session.begin() as sess:
            q = sess.query(Image.path, Image.timestamp).\
                filter(Image.location==None).all()
        return [FileMeta(path=Path(i[0]), date=i[1], country="Uknown", city="Unknown") for i in q]
    
    def find_photos_by_city_name(self, distance_km: int, name: str) -> list[Path]:
        with self.session.begin() as sess:
            q: list[tuple[str]] = sess.query(Image.path)\
                .join(City, Image.location.ST_DWithin(City.location, distance_km * 1000, True))\
                    .filter(City.name == name).all()
        return [Path(i[0]) for i in q]
    
    def find_photos_by_country_name(self, name: str) -> list[Path]:
        with self.session.begin() as sess:
            q = sess.query(Image.path)\
                .join(Country, Country.geometry.ST_Contains(Image.location))\
                    .filter(Country.name == name)
        return [Path(i[0]) for i in q]
     

if __name__ == "__main__":
    # import time
    # from core.image_paths import PhotoPaths
    # from core.image_metadata import MetadataExtractor

    # t = time.time()
    # p = PhotoPaths(["/media/storage/Photo"])
    # meta = MetadataExtractor(p)

    # api = DbApi()
    # api.add_photo_to_db(meta.metadata)
    # print(f"Completed in {time.time() - t} seconds")

    api = DbApi()
    # res = api.select_files_and_output(location=(-31.95, 115.86, 0), distance_km=50)
    # res = api.get_photo_country()
    
    res = api.find_photos_by_country_name('Russia')
    print(res)



from pathlib import Path
import datetime as dt
from dataclasses import dataclass, field


@dataclass
class FileMeta:
    path: str | Path
    date: dt.datetime
    country: str | None = field(default='')
    city: str | None = field(default='')


import time
from loguru import logger


def func_time(func):
    """Measure execution time of a method/function"""
    def wrapper(*args, **kwargs):
        t = time.time()
        res = func(*args, **kwargs)
        logger.info(f"{func.__name__} executed in {time.time() - t} s.")
        return res

    return wrapper



import json
from pathlib import Path

from loguru import logger
from geoalchemy2 import Geometry
from sqlalchemy.orm import declarative_base, relationship, sessionmaker
from sqlalchemy import (
    Column,
    ForeignKey,
    Integer,
    String,
    Float,
    DateTime,
    UniqueConstraint,
    Table,
    create_engine,
)

from gisterical.settings.settings import load_settings

SETTINGS = load_settings()

Base = declarative_base()
engine = create_engine(SETTINGS.conn_str)


image_objects = Table(
    "image_objects",
    Base.metadata,
    Column("image_id", ForeignKey("image.id")),
    Column("object_id", ForeignKey("object.id")),
    Column("match_accuracy", Float),
)


class Image(Base):
    __tablename__ = "image"

    id = Column(Integer, primary_key=True, autoincrement=True)
    path = Column(String)
    location = Column(Geometry("POINTZ"))
    timestamp = Column(DateTime)
    gps_accuracy = Column(Float)
    photo_direction = Column(Float)
    device_make = Column(String)
    device_model = Column(String)
    phash = Column(String)
    colorhash = Column(String)

    objects = relationship("Object", secondary=image_objects)

    unique_path = UniqueConstraint("path", name="unique_path")


class Object(Base):
    __tablename__ = "object"

    id = Column(Integer, primary_key=True, autoincrement=True)
    object_type = Column(String)
    object_name = Column(String)
    object_path = Column(String)


class Country(Base):
    __tablename__ = "country"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String)
    iso_code = Column(String)
    geometry = Column(Geometry)


class City(Base):
    __tablename__ = "city"

    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String)
    country = Column(String)
    location = Column(Geometry("POINT"))
    population = Column(Integer)


def populate_cities(session: sessionmaker):
    logger.info("Adding coordinates for world cities.")
    cities = Path(__file__).parent.parent / SETTINGS.cities_data
    with open(cities, "r") as f:
        with session.begin() as sess:
            for idx, row in enumerate(f.readlines()):
                r = row.split(",")
                
                try:
                    city = r[0].replace('"', "")
                except (TypeError, ValueError):
                    city = None
                    
                try:
                    country = r[4].replace('"', "")
                except (TypeError, ValueError):
                    country = None
                
                try:
                    pop = int(r[9].replace('"', ""))
                except (TypeError, ValueError):
                    pop = None

                try:
                    lat = float(r[2].replace('"', ""))
                    lon = float(r[3].replace('"', ""))
                    loc = f"POINT({lon} {lat})"
                except (TypeError, ValueError):
                    loc = None

                if idx >= 1:
                    new_city = City(
                        name=city,
                        country=country,
                        location=loc,
                        population=pop,
                    )
                    sess.add(new_city)
                    sess.flush()
            sess.commit()


def populate_countries(session: sessionmaker):
    logger.info("Adding country boundaries.")
    path = Path(__file__).parent.parent / SETTINGS.countries_data
    with open(path) as f:
        cnt = json.load(f)
    ft = cnt["features"]
    with session.begin() as sess:
        for f in ft:
            new_cnt = Country(
                name=f["properties"]["ADMIN"],
                iso_code=f["properties"]["ISO_A3"],
                geometry=json.dumps(f["geometry"]),
            )
            sess.add(new_cnt)
        sess.commit()


def create_schema():
    session = sessionmaker(engine)
    Base.metadata.create_all(engine)
    populate_cities(session)
    populate_countries(session)


if __name__ == "__main__":
    create_schema()



import os
import time
import warnings
import datetime as dt
from pathlib import Path
from typing import Any

from PIL import Image as PILImage
from exif import Image
import imagehash as imh
from attrs import define
from loguru import logger

from gisterical.core.image_paths import get_paths

warnings.filterwarnings('ignore', module='exif')

@define
class PhotoData:
    path: str
    latitude: float
    longitude: float
    altitude: float
    timestamp: dt.datetime
    gps_accuracy: float
    photo_direction: float
    camera_make: str
    camera_model: str
    phash: str | None = None
    colorhash: str | None = None


class MetadataExtractor:
    def __init__(self, image_paths: tuple[Path], hash_images: bool = False):
        logger.info("Collecting metadata from image files.")
        self.paths: tuple[Path] = image_paths
        self.__hash_images = hash_images
        self._raw_metadata = self.__raw_metadata()

    def __raw_metadata(self) -> dict[str, dict[str, Any]]:
        # this is bloody ugly but I' building the dict instead of
        # using Image object because the latter eats too much memory
        # and has caused crashes in the past. With this memory 
        # issues are solved.
        data: dict[str, dict[str, Any]] = {}
        for p in self.paths:
            with open(p, "rb") as f:
                tmp = Image(f)
                try:
                    lat = tmp.get('gps_latitude')
                    lat_ref = tmp.get('gps_latitude_ref')
                    lon = tmp.get('gps_longitude')
                    lon_ref = tmp.get('gps_longitude_ref')
                    alt = tmp.get('gps_altitude')
                except (AttributeError, KeyError):
                    lat = lon = alt = None
                    lat_ref = lon_ref = None
                    
                try:
                    date = tmp.get('datetime_original') or '1900:1:1 00:00:00'
                except (AttributeError, KeyError):
                    date = '1900:1:1 00:00:00'
                    
                try:
                    hor_pos = tmp.get('gps_horizontal_positioning_error')
                    direct = tmp.get('gps_img_direction')
                except (AttributeError, KeyError):
                    hor_pos = direct = -999
                    
                try:
                    mk = tmp.get('make')
                    mod = tmp.get('model') 
                except (AttributeError, KeyError):
                    mk = "unknown device"
                    mod = "unknown model"
                
                
                data[str(p)] = {'gps_latitude': lat,
                    'gps_latitude_ref': lat_ref,
                    'gps_longitude': lon,
                    'gps_longitude_ref': lon_ref,
                    'gps_altitude': alt,
                    'datetime_original': date,
                    'gps_horizontal_positioning_error': hor_pos,
                    'gps_img_direction': direct,
                    'make': mk,
                    'model': mod                 
                }
        return data

    @property
    def metadata(self) -> list[PhotoData]:
        res: list[PhotoData] = []
        for cnt, (pth, dic) in enumerate(self._raw_metadata.items(), start=1):
            # logger.info(f"Processing image {cnt}/{len(self._raw_metadata)}")
            try:
                lat = self._convert_coords_to_decimal(dic['gps_latitude'], dic['gps_latitude_ref'])
                lon = self._convert_coords_to_decimal(dic['gps_longitude'], dic['gps_longitude_ref'])
                alt = dic['gps_altitude']
            except (AttributeError, KeyError):
                lat = lon = alt = -999

            # when getting dates so images (like those sent through WhataApp, Telegram etc)
            # won't have correct date recorded so instead we take the last modified date
            # and select it instead            
            try:                
                timestamp = dt.datetime.strptime(dic['datetime_original'], "%Y:%m:%d %H:%M:%S")
            except (AttributeError, KeyError, ValueError):
                ts = os.path.getmtime(pth)
                timestamp = dt.datetime.utcfromtimestamp(ts)
            else:
                # if the image has datetime recorded we check whether it is in the future
                # (due to incorrect camera setup and such) or whether it is before 1/1/1970
                # and if either of these is true then we take modified date still
                if timestamp > dt.datetime.now() or timestamp < dt.datetime(1970, 1, 1):
                    ts = os.path.getmtime(pth)
                    timestamp = dt.datetime.utcfromtimestamp(ts)                

            gps_accuracy = dic['gps_horizontal_positioning_error']
            photo_direction = dic['gps_img_direction']
            camera_make = dic['make']
            camera_model = dic['model']

            if self.__hash_images:
                logger.debug("Calculating image hashes.")
                t = time.time()
                im = PILImage.open(str(pth))
                phash = str(imh.phash(im))
                chash = str(imh.colorhash(im))
                print(f"Hashing image took {time.time() - t} seconds")
            else:
                phash = chash = None

            res.append(
                PhotoData(
                    path=pth,
                    latitude=lat,
                    longitude=lon,
                    altitude=alt,
                    timestamp=timestamp,
                    gps_accuracy=gps_accuracy,
                    photo_direction=photo_direction,
                    camera_make=camera_make,
                    camera_model=camera_model,
                    phash=phash,
                    colorhash=chash,
                )
            )
        return res

    def _convert_coords_to_decimal(self, coords: tuple[float, ...], ref: str) -> float:
        """Covert a tuple of coordinates in the format (degrees, minutes, seconds)
        and a reference to a decimal representation.

        Args:
            coords (tuple[float,...]): A tuple of degrees, minutes and seconds
            ref (str): Hemisphere reference of "N", "S", "E" or "W".

        Returns:
            float: A signed float of decimal representation of the coordinate.
        """
        if ref.upper() in {"W", "S"}:
            mul = -1
        elif ref.upper() in {"E", "N"}:
            mul = 1
        else:
            msg = f"Incorrect hemisphere reference. Expecting one of 'N', 'S', 'E' or 'W', got {ref} instead."
            logger.debug(msg)
            raise ValueError(msg)
        return mul * (coords[0] + coords[1] / 60 + coords[2] / 3600)


if __name__ == "__main__":
    import sys
    import time

    t = time.time()
    p = get_paths(["/media/storage/Photo"])
    meta = MetadataExtractor(p, hash_images=True)
    r = meta.metadata
    elapsed = time.time() - t

    print(f"Length: {len(r)} -- Memory: {sys.getsizeof(r)} -- Elapsed: {elapsed}")



from __future__ import annotations

from typing import Any
from pathlib import Path
from shutil import copyfile
from dataclasses import dataclass, field

from gisterical.util import FileMeta


def filter_year(data: list[FileMeta]) -> set[int]:
    return {i.date.year for i in data}


def filter_month(data: list[FileMeta]) -> set[int]:
    return {i.date.month for i in data}


def filter_day(data: list[FileMeta]) -> set[int]:
    return {i.date.day for i in data}


def filter_country(data: list[FileMeta]) -> set[str]:
    return {i.country for i in data}


def filter_city(data: list[FileMeta]) -> set[str]:
    return {i.city for i in data}


condition_dict = {
    "Y": {"fun": filter_year, "id": 1},
    "m": {"fun": filter_month, "id": 1},
    "d": {"fun": filter_day, "id": 1},
    "C": {"fun": filter_country, "id": 2},
    "c": {"fun": filter_city, "id": 3},
}


@dataclass
class Node:
    folder: Path
    metadata: list[FileMeta] = field(default_factory=list)
    conditions: list[str] = field(default_factory=list)

    @property
    def children(self) -> list["Node"] | None:
        if len(self.conditions) == 0:
            return None
        current_condition = self.conditions[0]
        fun = condition_dict[current_condition]["fun"]

        fnames: list[int | str] = fun(self.metadata)
        children_data: dict[str, dict[str, Any]] = {str(n): {} for n in fnames}
        for f in fnames:
            children_data[str(f)]["folder"] = self.folder / str(f)
            children_data[str(f)]["conditions"] = self.conditions[1:]
            if current_condition == "C":
                children_data[str(f)]["metadata"] = [i or "Unknown_country" for i in self.metadata if i.country == f]

            elif current_condition == "Y":
                children_data[str(f)]["metadata"] = [i for i in self.metadata if i.date.year == f]
            elif current_condition == "c":
                children_data[str(f)]["metadata"] = [i or "Unknown_city" for i in self.metadata if i.city == f]


            elif current_condition == "d":
                children_data[str(f)]["metadata"] = [i for i in self.metadata if i.date.day == f]
            elif current_condition == "m":
                children_data[str(f)]["metadata"] = [i for i in self.metadata if i.date.month == f]
        return [Node(**i) for i in children_data.values()]


def traverse(
    root: Node, folders: dict[Path, list[str | Path]] | None = None
) -> dict[Path, list[str | Path]]:
    # build a dictionary with folderPath as the key and list of file paths as value
    folders = folders or {}
    r = root.children
    if not r or len(root.conditions) == 0:
        if root.folder not in folders:
            # for leaf node add folder name and corresponding files
            folders[root.folder] = [i.path for i in root.metadata]
    else:
        for child in r:
            if isinstance(child, Node):
                par = child.folder.parent
                # we only want the data from leaf nodes so
                # if the record for the current folder's parent
                # already exists we delete it from the dictionary
                # and add the current folder data
                if par in folders:
                    del folders[par]
                folders[child.folder] = [i.path for i in child.metadata]
                traverse(child, folders)
    return folders


def make_folder(path: Path, children: list[Path] = None) -> None:
    if path.exists():
        return
    children = children or []
    if path.parent.is_dir():
        # print(f'Creating folder {path}')
        path.mkdir()
        if not children:
            return
        return make_folder(children[0])

    else:
        # print(f'Folder {path} does not exist')
        children.append(path)
        return make_folder(path.parent, children)


def populate_folder_structure(files: dict[Path, list[str | Path]]):
    for target_fol, original_files in files.items():
        for f in original_files:
            target_file = target_fol / Path(f).name
            copyfile(str(f), str(target_file))


if __name__ == "__main__":
    pass
    # from database.db_api import DbApi
    # COND = ["Y", "m", ]

    # api = DbApi()
    # d = api.get_photo_path_date()
    
    # n = Node(Path('temp'), metadata=d, conditions=COND)
    
    # tree = traverse(n)
    
    # for leaf in tree:
    #     make_folder(leaf)
        
    # move_files(tree)




from pathlib import Path

import cv2
import numpy as np
import face_recognition as fc

# from core.image_paths import PhotoPaths

fol = Path("/media/storage/Photo/Photo/New Zealand")
# file = "/media/storage/Photo/Photo/New Zealand/DSC_7283.jpg"
files = list(fol.glob("*.jpg"))

for i, file in enumerate(files):
    print(f"Processing file {i+1}/{len(files)}")
    image = fc.load_image_file(file)
    face_locs: list[tuple[int, int, int, int]] = fc.face_locations(
        image, number_of_times_to_upsample=1
    )

    f = cv2.imread(str(file))

    if len(face_locs):
        for idx, r in enumerate(face_locs):
            x1 = np.maximum(0, r[0] - 100)
            x2 = np.minimum(image.shape[0], r[2] + 100)
            y1 = np.maximum(0, r[3] - 100)
            y2 = np.minimum(image.shape[1], r[1] + 100)

            f1 = f[x1:x2, y1:y2]
            try:
                cv2.imwrite(f"/home/pavel/Pictures/temp/{str(file.name)}-{idx}.png", f1)
            except Exception as e:
                print(e)



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name='GISterical',
    version='0.5.8.19',
    license='MIT',
    author="Pavel Cherepanskiy",
    author_email='pavelcherepansky@gmail.com',
    packages=find_packages('src'),
    package_dir={
        'gisterical': 'src/gisterical',
        },
    package_data={
        'gisterical': ['data/countries.geojson', 'data/worldcities.csv']
    },
    url='https://github.com/pavelcherepan/gisterical',
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords='example project',
        classifiers=[
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
    ],
    python_requires=">=3.9, <4",
    install_requires=[
          'loguru', 
          'SQLAlchemy>=1.4', 
          'GeoAlchemy2>=0.12',
          'exif',
          'attrs',
          'ImageHash',
          'Pillow',
          'psycopg2-binary>=2.8'
      ],
    entry_points={
        "console_scripts": [
            "gisterical=gisterical:main",
        ],
    },
)


from pathlib import Path


def get_paths(
    folders: list[str | Path],
    search_extensions: tuple[str, ...] = (
        ".jpg",
        ".jpeg",
        ".tif",
        ".tiff",
        ".bmp",
        ".gif",
        ".png",
    ),
) -> tuple[Path]:
    all_files: list[Path] = []
    for fol in folders:
        if isinstance(fol, str):
            fol = Path(fol)
        
        # on occasion extensions can be in upper case so adding
        # a second array containing uppercase extensions
        ext = list(search_extensions)
        extensions = ext + [i.upper() for i in ext]
        
        for e in extensions:
            all_files.extend(list(fol.rglob(f"*{e}")))
    return tuple(all_files)


if __name__ == "__main__":
    p = get_paths(["/media/storage/Photo/Photo/Phone",])
    for i in p:
        print(i)



import argparse
from time import time
from pathlib import Path
from loguru import logger
from shutil import copyfile

from gisterical.core.image_metadata import MetadataExtractor
from gisterical.core.image_paths import get_paths
from gisterical.core.create_folder_structure import Node, traverse, populate_folder_structure, make_folder
from gisterical.database.db_api import DbApi
from gisterical.database.schema import create_schema
from gisterical.settings.settings import load_settings, update_settings
from gisterical.util.file_meta import FileMeta


SETTINGS = load_settings()
api = DbApi()
parser = argparse.ArgumentParser(
    description="Sort images based on date and/or location using a local PostGIS database to store metadata.",
)

# give option of using either a long or short labels
inp = parser.add_mutually_exclusive_group(required=False)
out = parser.add_mutually_exclusive_group(required=False)
nam = parser.add_mutually_exclusive_group(required=False)

parser.add_argument(
    "--set-connection", 
    action="store_true", 
    help="Perform intial setup", 
    default=False, 
    required=False
)
parser.add_argument(
    "--setup", 
    action="store_true", 
    help="Perform intial setup", 
    default=False, 
    required=False
)
parser.add_argument(
    "--add-folder", 
    action="store_true", 
    help="Add photos to the database. Input folder has to be provided.",
    default=False,
    required=False
)
parser.add_argument(
    "--sort",
    action="store",
    type=str,
    help="Sort images and output to a folder",
    nargs="+",
    default=False, 
    required=False
)

inp.add_argument("--input", action="store", type=str, help="Input folder for initial setup")
inp.add_argument("-i", action="store", type=str, help="Input folder for initial setup")

out.add_argument("--output", action="store", type=str, help="Output folder for sorting and search operations")
out.add_argument("-o", action="store", type=str, help="Output folder for sorting and search operations")

parser.add_argument("--find-by-city", action="store", type=str, 
                    help="Find photos taken within certain distance from a specific city and output to target folder. " 
                    "Output folder, city name, and distance parameters have to be provided.")
parser.add_argument("--find-by-country", action="store", type=str, 
                    help="Find photos taken within a country and output to target folder. "
                    "Country name and output folder parameters need to be provided.")
nam.add_argument("--name", action="store", type=str, help="A city or country name to search in photos.")
nam.add_argument("-n", action="store", type=str, help="A city or country name to search in photos.")

parser.add_argument("--distance", action="store", type=int, help="Maximum match distance to between a photo location and a city.")
 
parser.add_argument(
    "--hash",
    help="Calculate image hashes during setup and store in the database",
    action="store_true",
)

args = parser.parse_args()


def perform_initial_setup(source_folder: str):
    """Perform initial set-up of the database for which we create 
    the base schema, extract the necessary metadata extracted from images,
    and populate image data along with the country and city data into
    the db.

    Args:
        source_folder (str): A string of the source folder containing images.
    """
    do_hash = bool(args.hash)
    create_schema()
    p = get_paths([source_folder])
    meta = MetadataExtractor(p, hash_images=do_hash)
    api.add_photo_to_db(meta.metadata)


def check_flags(args: argparse.Namespace) -> tuple[list[str], int, Path]:
    """Check positional flags and validate.

    Args:
        args (argparse.Namespace): A Namespace object containing all passed arguments

    Raises:
        ValueError: raised if unrecognised arguments are passed; output folder is missing;
            distance argument not present when "c" flag is included.

    Returns:
        tuple[list[str], int, Path]: A tuple consisting of a list of individual positional params, 
            integer of distance to find nearest city in kilometres and the output path.
    """
    inp = list(args.sort[0]) if len(args.sort) == 1 else args.sort
    v = set(inp).intersection({'C', 'c', 'Y', 'm', 'd'})

    if not args.output and not args.o:
        raise ValueError('Output folder parameter required when sorting!')
    p = args.output or args.o
    pth = Path(p)

    if len(v) != len(inp):
        logger.exception(f'Some of the input arguments {inp} not recognised. '
                        f'Accepted flags are "C", "c", "Y", "m" and "d".')
        raise ValueError('Positional arguments not recognised!')
    if "c" in inp and not args.distance:
        raise ValueError("Distance argument (--distance) required when sorting by city.")
    return inp, args.distance , pth 


def _parse_positional_args(input_args: argparse.Namespace) -> tuple[Path, list[FileMeta], list[str]]:
    """Parse positional flags used for sorting and decide which specific
    database api method to run to get the data necessary to build a tree-like
    folder structure.

    Args:
        input_args (argparse.Namespace): A Namespace object with input arguments.

    Returns:
        tuple[Path, list[FileMeta], list[str]]: A tuple consisting of the output path location,
            a list of FileMeta objects representing required file metadata and the list of
            validated sorting flags.
    """
    sorted_flags, distance, out_path = check_flags(input_args) 
    if len({"C", "c"}.intersection(sorted_flags)) == 2:
        # a lot of photos don't have location data but often you'd still want
        # to sort them by date. If you do a spatial join then these photos will
        # be missed so we run another query on the db where we extract all
        # photos with missing location information. This is computationally
        # very cheap so there's little benefit to not doing this.
        data_mis = api.get_photo_no_location()
        data_loc = api.get_photo_city_country(distance)
        data = data_loc + data_mis
    elif "c" in sorted_flags:
        data_mis = api.get_photo_no_location()
        data_loc = api.get_photos_by_city(distance)
        data = data_loc + data_mis
    elif "C" in sorted_flags:
        data_mis = api.get_photo_no_location()
        data_loc = api.get_photo_country()
        data = data_mis + data_loc
    else:
        data = api.get_photo_path_date()
    return out_path, data, sorted_flags


def run_sort_task(input_args: argparse.Namespace):
    """Run the sorting task. For this use the helper function to
    get the required data, then using metadata extracted from db
    build an m-ary tree of folder structure, traverse the tree to
    build the dataset and finally create necessary folders and
    move files to new locations.

    Args:
        input_args (argparse.Namespace): A Namespace object with input arguments.
    """
    out_path, data, sorted_flags = _parse_positional_args(input_args)    
    n = Node(out_path, data, sorted_flags)
    tree = traverse(n)
        
    for leaf in tree:
        make_folder(leaf)
    
    populate_folder_structure(tree)  
 
 
def copy_files(paths: list[Path], target_folder: Path):
    if not target_folder.exists():
        make_folder(target_folder)
    for f in paths:
        target_file = target_folder / f.name
        copyfile(str(f), str(target_file))  
        

def _validate_search_inputs(args: argparse.Namespace) -> Path:
    if not args.output and not args.o:
        raise ValueError('Output folder must be provided for search operations.')
    t = args.output or args.o
    out = Path(t)
    if args.find_by_city and not args.distance:
        raise ValueError('Distance needs to be provided to search by city name.')
    return out
    
        
def main():   
    t = time() 
    if args.set_connection:
        update_settings()
    elif args.setup and (args.input or args.i):
        perform_initial_setup(args.input or args.i)
    elif args.sort:
        run_sort_task(args)
    elif args.find_by_city:
        out = _validate_search_inputs(args)
        paths = api.find_photos_by_city_name(args.distance, args.find_by_city)
        copy_files(paths, out)
    elif args.find_by_country:
        out = _validate_search_inputs(args)
        paths = api.find_photos_by_country_name(args.find_by_country)
        copy_files(paths, out)       
    elif args.add_folder:
        if not args.input and not args.i:
            raise ValueError("Input folder must be provided.")
        fol = args.input or args.i
        p = get_paths([fol])
        meta = MetadataExtractor(p, hash_images=args.hash)
        api.add_photo_to_db(meta.metadata)
    logger.info(f'Successfully completed in {time() - t} seconds.')        


if __name__ == "__main__":
    main()
